Program mpi_scatterv2
!============================================================!
! Programmer: Spiros Vellas@Texas A&M                        !
!                                                            !
! Illustrates use of MPI_SCATTERV in carrying matrix-vector  !
! multiplication. Strips of consecutive columns of matrix A  !
! are distributed to all processes. Each of these carries out!
! a part of the linear vector sum:                           !
!                                                            !
!   A*b = b1*a1 + b2*a2 + ... + bn*an = c                    !
!                                                            !
! where the a's are the columns of A = [ a1 a2 ... an ]  and !
! the b's are the (scalar) components of the b column vector !
!                                                            !
!	b = [ b1 b2 ... bn ]'  ("'" stands for transpose)    !
!                                                            !
!============================================================!
USE STDSPECS
USE MPI
integer id, ierr, n, rank, ncols, np, L1, L2, root
integer, allocatable :: sendcnt(:), offsets(:)
real (kind=DP), allocatable :: A(:, :), alocal(:,:)
real (kind=DP), allocatable :: b(:), ab(:), c(:), c2(:)
real (kind=DP) tol, e_tm
!--------------------------------------------------------------
call MPI_INIT(ierr)
call MPI_COMM_RANK(MPI_COMM_WORLD, rank, ierr)
call MPI_COMM_SIZE(MPI_COMM_WORLD, np, ierr)
!
root = 0
!
if ( rank == root ) then
   print *, 'Enter dimension size - '; read(unit=*,fmt=*) n
   ALLOCATE( A(1:n,1:n), b(1:n), c(1:n), c2(1:n), sendcnt(0:np-1), &
                                                  offsets(0:np-1) )
   call random_number(A); call random_number(b) ! Fill them with rands
!
   do id=0, np-1
      call BLOCK_MAP(1, n, np, id, L1, L2) ! Block-Map iters
      sendcnt(id) = ( L2 - L1 + 1 ) * n       ! Space id will need
      if ( id == 0 ) then
         offsets(0) = 0
      else
         offsets(id) = offsets(id-1) + sendcnt(id-1)
      end if
   end do
endif
!
call MPI_BCAST(n, 1, MPI_INTEGER, root, MPI_COMM_WORLD, ierr)
If ( rank /= root ) ALLOCATE(  b(1:n) )
!
call MPI_BCAST(b, n, MPI_REAL8, root, MPI_COMM_WORLD, ierr)
!
call BLOCK_MAP(1, n, np, rank, L1, L2) ! Block-Map iters
ncols = L2 - L1 + 1
!
ALLOCATE( alocal(1:n, 1:ncols), ab(1:n) ); alocal = 0.0D0
!
call mpi_scatterv(A, sendcnt, offsets, MPI_REAL8, alocal, &
                  n*ncols, MPI_REAL8, root, MPI_COMM_WORLD, ierr)
ab(1:n) = 0.0D0
do i=L1, L2
   ab = ab + b(i) * alocal(1:n, i-L1+1)
end do
!
call MPI_REDUCE(ab, c, n, MPI_REAL8, MPI_SUM, root, MPI_COMM_WORLD, ierr)
!
if ( rank == root ) then
   c2 = matmul(A,b)             ! F90 intrinsic: c2=A*b
!                                 Confirm correctness
   write(6,fmt='(/" *** ",I4," x",I4, &
&         " Matrix-Vector Multiplication ***",/)' ) n,n
   write(6,fmt='("   -- MATMUL --   -- SCATTERV2 --")' )
   write(6,fmt='(4X,F9.4,6X,F9.4)') (c2(i), c(i), i=1, n)
end if
!
call MPI_FINALIZE(ierr)
end program mpi_scatterv2


subroutine block_map(n1,n2, &
         nprocs, myid, l1, l2)

Integer n1, n2, nprocs, &
        myid, l1,l2
integer block, rem
block = (n2-n1+1)/nprocs
rem = mod(n2-n1+1, nprocs)
if (myid < rem) then
    block = block+1
    l1 = n1+myid*block
else
    l1 = n1+rem+block*myid
end if
l2 = l1+block-1
end subroutine block_map

!====================================================================
