/*
This program is a re-wirte in C of the Fortran poisson_1d.f90.
See the annotations in poisson_1d.f90 for more information about what the code does. 

This program uses MPI 1d virtual topology to create a 1D cartesian grid. 
The neighbour information will be retrieved using MPI function calls.
Programmer: Ping Luo
Supercomputing Facility
Texas A&M University
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include <omp.h>

#define PI 3.1415926535897932
#define MIN(a, b) (((a)<(b))?(a):(b))
#define IND(i, j, n) ((i)*(n)+(j))

void create_cart1d(int nprocs, int *iprevid, int *inextid, MPI_Comm *comm1d, int *myid);
void block_map(int n, int nprocs, int myid, int *s, int *e);
void init1d_domain(int n, int s, int e, double *f, double *u, double *unew);
void sweep1d(double *u, double *f, int n, int s, int e, double *unew);
void exchng1d(MPI_Comm comm1d, int iprevid, int inextid, int n, int s, int e, double *u);
double diff(double *u, double *unew, int n, int s, int e);
void compare_solns(double *u, int s, int e, int n);

double sweep_t = 0.0;
int main(int argc, char **argv){

    int n;  // mesh size nxn
    MPI_Comm comm1d;
    int myid, nprocs, s, e, iprevid, inextid;

    double *f, *u, *unew;
    int iter = 0, max_iter;
    double eps = 0.000000001, diffnorm = 1.0, mydiff = 0.0;
 
    double t1, t2;

    int i;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

    if (myid == 0){
        printf("Enter the mesh size - ");
        scanf("%d", &n);
        printf("Enter maximum number of iterations - ");
        scanf("%d", &max_iter);
    }

    MPI_Bcast(&n, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&max_iter, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    create_cart1d(nprocs, &iprevid, &inextid, &comm1d, &myid);
    block_map(n, nprocs, myid, &s, &e);

    // for simplicity, we create f the same size as u and unew
    // the domain is split along x axis
    f    = (double *)calloc((e-s+3)*(n+2), sizeof(double));
    u    = (double *)calloc((e-s+3)*(n+2), sizeof(double));
    unew = (double *)calloc((e-s+3)*(n+2), sizeof(double));

    if (f==NULL || u==NULL || unew==NULL){
        printf("Memory allocation error\n");
        MPI_Abort(MPI_COMM_WORLD, -1);
    }

    init1d_domain(n, s, e, f, u, unew);

    MPI_Barrier(MPI_COMM_WORLD);

    t1 = MPI_Wtime();
    while (iter < max_iter && diffnorm > eps){
        sweep1d(u, f, n, s, e, unew);
        exchng1d(comm1d, iprevid, inextid, n, s, e, unew);
        sweep1d(unew, f, n, s, e, u);
        exchng1d(comm1d, iprevid, inextid, n, s, e, u);
        mydiff = diff(unew, u, n, s, e);
        MPI_Allreduce(&mydiff, &diffnorm, 1, MPI_DOUBLE, MPI_SUM, comm1d);
        iter += 2;
    }
    t2 = MPI_Wtime();


    if (myid == 0){
        printf("Wall clock time: %10.4f seconds\n", t2-t1);
        if (diffnorm <= eps){
            printf("Converged: iter=%d, diffnorm=%15.14f\n", iter, diffnorm);
        }else{
            printf("Not converged: iter=%d, diffnorm=%15.14f\n", iter, diffnorm);
        }
    }

    for (i=0; i<nprocs; i++){
        if (myid == i) {
            printf("%d\n", myid);
            compare_solns(u, s, e, n);
        }
        MPI_Barrier(comm1d);
    }

    free(f);
    free(u);
    free(unew);

    MPI_Comm_free(&comm1d);
    MPI_Finalize();

}


void
create_cart1d(int nprocs, int *iprevid, int *inextid, MPI_Comm *comm1d, int *myid)
{
    int ndim = 1;
    int dims[ndim];
    int isperiodic[ndim];
    int reorder = 1;

    dims[0] = nprocs;
    isperiodic[0] = 0;
    MPI_Cart_create(MPI_COMM_WORLD, ndim, dims, isperiodic, reorder, comm1d);
    MPI_Comm_rank(*comm1d, myid);
    MPI_Cart_shift(*comm1d, 0, 1, iprevid, inextid);
}

void
block_map(int n, int nprocs, int myid, int *s, int *e)
{
    if (myid < n%nprocs){
        *s = (n/nprocs+1)*myid + 1;
        *e = *s + n/nprocs;
    }else{
        *e = n - n/nprocs*(nprocs-myid-1);
        *s = *e - n/nprocs + 1;
    }
}

void
init1d_domain(int n, int s, int e, double *f, double *u, double *unew)
{
    double h = 1.0/(double)(n+1), x, y;
    double pisq = PI*PI;
    int block_size = e-s+3;
    int i, j;

    // initialize f
    for (i=1; i<block_size-1; i++){
        x = (i+s-1)*h;
        for (j=1; j<=n; j++){
            y = j*h;
            f[IND(i,j,(n+2))] = (x*x+y*y)*cos(x*y)-cos(PI*x);
        } 
    }

    // initialize left border of u and unew
    // x = 0
    if (s == 1){
        i = 0;
        for (j=0; j<=n+1; j++){
            u[IND(i,j,(n+2))] = 1.0/pisq-1.0;
            unew[IND(i,j,(n+2))] = u[IND(i,j,(n+2))];
    }
    }

    // initialize right border of u and unew
    // x = 1
    if (e == n){
    i = e-s+2;
    for (j=0; j<=n+1; j++){
        y = j*h;
        u[IND(i,j,(n+2))] = -(1.0/pisq+cos(y));
        unew[IND(i,j,(n+2))] = u[IND(i,j,(n+2))];
    }
    }

    // initialize upper border of u and unew
    // y = 1
    j = n+1;
    for (i=0; i<block_size; i++){
        x = (i+s-1)*h;
        u[IND(i,j,(n+2))] = (cos(PI*x) - pisq*cos(x))/pisq;
        unew[IND(i,j,(n+2))] = u[IND(i,j,(n+2))];
    }
 
    // initialize lower border of u and unew
    // y = 0
    j = 0;
    for (i=0; i<block_size; i++){
        x = (i+s-1)*h;
        u[IND(i,j,(n+2))] = (cos(PI*x) - pisq)/pisq;
        unew[IND(i,j,(n+2))] = u[IND(i,j,(n+2))];
    }
}

void
sweep1d(double *u, double *f, int n, int s, int e, double *unew)
{
    double hsq = 1.0/(double)((n+1)*(n+1));
    int i, j;
    int block_size = e-s+1;

#pragma omp parallel for private(i) 
    for (i=1; i<=block_size; i++){
        for (j=1; j<n+1; j++){
            unew[IND(i,j,(n+2))] = (u[IND((i-1),j,(n+2))] +
                                    u[IND((i+1),j,(n+2))] +
                                    u[IND(i,(j-1),(n+2))] +
                                    u[IND(i,(j+1),(n+2))] - 
                                    hsq*f[IND(i,j,(n+2))]) * 0.25;
        }
    }
}

double 
diff(double *u, double *unew, int n, int s, int e)
{
    double diffnorm = 0.0;
    int block_size = e-s+1;
 
    int i, j;
 
#pragma omp parallel for schedule(static) private(i) reduction(+:diffnorm)
    for (i=1; i<=block_size; i++)
        for (j=1; j<n+1; j++){
            diffnorm += (unew[IND(i,j,(n+2))]-u[IND(i,j,(n+2))]) * (unew[IND(i,j,(n+2))]-u[IND(i,j,(n+2))]);
        } 

    return diffnorm;
}

void 
exchng1d(MPI_Comm comm1d, int iprevid, int inextid, int n, int s, int e, double *u)
{
    MPI_Status status;
    MPI_Request reqs[4];

/*
    MPI_Send(&u[IND(1,1,(n+2))], n, MPI_DOUBLE, iprevid, 0, comm1d);
    MPI_Recv(&u[IND((e-s+2),1,(n+2))], n, MPI_DOUBLE, inextid, 0, comm1d, &status);
    MPI_Send(&u[IND((e-s+1),1,(n+2))], n, MPI_DOUBLE, inextid, 1, comm1d);
    MPI_Recv(&u[IND(0,1,(n+2))], n, MPI_DOUBLE, iprevid, 1, comm1d, &status);
*/

    // Non-blocking send and receive
    MPI_Isend(&u[IND(1,1,(n+2))], n, MPI_DOUBLE, iprevid, 0, comm1d, &reqs[0]);    
    MPI_Irecv(&u[IND((e-s+2),1,(n+2))], n, MPI_DOUBLE, inextid, 0, comm1d, &reqs[1]);
    MPI_Isend(&u[IND((e-s+1),1,(n+2))], n, MPI_DOUBLE, inextid, 1, comm1d, &reqs[2]);
    MPI_Irecv(&u[IND(0,1,(n+2))], n, MPI_DOUBLE, iprevid, 1, comm1d, &reqs[3]);
    MPI_Waitall(4, reqs, MPI_STATUSES_IGNORE);
/*
    // Combined sendrecv
    MPI_Sendrecv(&u[IND(1,1,(n+2))], n, MPI_DOUBLE, iprevid, 0, &u[IND((e-s+2),1,(n+2))], n, MPI_DOUBLE, inextid, 0, comm1d, MPI_STATUS_IGNORE);
    MPI_Sendrecv(&u[IND((e-s+1),1,(n+2))], n, MPI_DOUBLE, inextid, 1, &u[IND(0,1,(n+2))], n, MPI_DOUBLE, iprevid, 1, comm1d, MPI_STATUS_IGNORE);
*/
}

void 
compare_solns(double *u, int s, int e, int n)
{
    double h = 1.0/(double)(n+1), PISQ=PI*PI;
    int i, j;

    for (i=s; i<=e; i+=n/MIN(n,5)){
        printf("=== Exact Solution        Jacobi Iteration ===\n");
        for (j=1; j<n+1; j+=n/MIN(n,5)){
            printf("%d %d %16.15f               %16.15f\n", i, j, (cos(PI*i*h)/PISQ)-cos(j*h*i*h), u[(i-s+1)*(n+2)+j]);
        }
    }
}
